--[[--------------------------------------------------------------------
	LOIHInterrupt
------------------------------------------------------------------------

	Interrupt manager for guild <Lords of Ironhearts> of Arathor EU.
	(c) 2015- Sanex

	Creates 1-5 bars showing availability of players' interrupt spells.
	Players with green bars saying 'Ready' are able to interrupt.
	Players with red bar with countdown have their interrupt spell on CD
	until the timer runs out and the bar turns green again.

	Interrupters can be easily configured via GUI and interrupt lists
	can be shared to other users of the addon as well printed to party
	or raid chat. Addon can also be set to announce when player
	interrupts spell.

	Interrupt lists are saved per character, but other settings are
	shared between characters.

	Thanks to Bloodline for suggesting interrupt announcements feature.

------------------------------------------------------------------------

	Known issues/bugs:
	- Warlock interrupt spells

	Todo:
	- 

------------------------------------------------------------------------

	-- Globals:
	LOIHInterruptDB
		Local referer 'cfg'
		Used to save settings, shared between characters.
		XPos and YPos numeric, rest boolean.

	LOIHInterruptCharDB
		Local referer 'db'
		Used to save list of interrupters, saved per character.
		String, can be empty strings.

	SLASH_LOIHINTERRUPT1
	SLASH_LOIHINTERRUPT2
		No local refrers
		Used for slash commands.

	All other globals are leaks and should be reported to maintainer.

--------------------------------------------------------------------]]--

local ADDON_NAME, private = ...
local _G = _G
local GetTime = _G.GetTime
local math = _G.math
local pairs = _G.pairs
local UnitHealth = _G.UnitHealth
local SetWidth = _G.SetWidth
local GetWidth = _G.GetWidth
local SetFormattedText = _G.SetFormattedText
local SetColorTexture = _G.SetColorTexture
local SetTexture = _G.SetTexture
local unpack = _G.unpack
local SetText = _G.SetText
local IsInGroup = _G.IsInGroup
local SendChatMessage = _G.SendChatMessage
local format = _G.format

local AceGUI = LibStub("AceGUI-3.0")

local db, cfg
local config, tab
local playerName, playerClass
local deadCheck
local openSetup = false
local combatslash = false
local interruptMsg = _G.INTERRUPTED.." |cff71d5ff|Hspell:%d:|h[%s]|h|r"
local defaultIcon = "Interface\\Icons\\INV_Misc_QuestionMark"
local Interrupters = {}
local deadWatch = {}
local partyTable = {}
local greenBar = { 0, .75, 0, .75 }
local redBar = { .75, 0, 0, .75 }
local blackBar = { 0, 0, 0, .75 }
local defaultDB = {
	debugmode = false,
	XPos = 350,
	YPos = 100,
	solo = false,
	party = true,
	raid = true,
	combatonly = true,
	announce = false,
	partyFill = true,
}

------------------------------------------------------------------------

local function Debug(text, ...)
	if not db or not cfg.debugmode then return end

	if text then
		if text:match("%%[dfqsx%d%.]") then
			(DEBUG_CHAT_FRAME or ChatFrame3):AddMessage("|cffff9999"..ADDON_NAME..":|r " .. format(text, ...))
		else
			(DEBUG_CHAT_FRAME or ChatFrame3):AddMessage("|cffff9999"..ADDON_NAME..":|r " .. strjoin(" ", text, tostringall(...)))
		end
	end
end

local function Print(text, ...)
	if text then
		if text:match("%%[dfqs%d%.]") then
			DEFAULT_CHAT_FRAME:AddMessage("|cffffcc00".. ADDON_NAME ..":|r " .. format(text, ...))
		else
			DEFAULT_CHAT_FRAME:AddMessage("|cffffcc00".. ADDON_NAME ..":|r " .. strjoin(" ", text, tostringall(...)))
		end
	end
end

local function initDB(db, defaults) -- This function copies values from one table into another:
	if type(db) ~= "table" then db = {} end
	if type(defaults) ~= "table" then return db end
	for k, v in pairs(defaults) do
		if type(v) == "table" then
			db[k] = initDB(db[k], v)
		elseif type(v) ~= type(db[k]) then
			db[k] = v
		end
	end
	return db
end

------------------------------------------------------------------------

local f = CreateFrame("Frame")
f:RegisterEvent("ADDON_LOADED")
f:SetScript("OnEvent", function(self, event, ...)
	return self[event] and self[event](self, event, ...)
end)

local BarFrame = CreateFrame("Frame", "LOIHInterruptFrame", _G.UIParent)
BarFrame:SetWidth(125)
BarFrame:SetHeight(80)
BarFrame:SetScript("OnShow", function(self)
	-- Listen to CLEU when visible
	f:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end)
BarFrame:SetScript("OnHide", function(self)
	-- Stop listening to CLEU when hidden
	f:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end)

local bg = BarFrame:CreateTexture(nil, "ARTWORK")
bg:ClearAllPoints()
bg:SetAllPoints()
bg:SetColorTexture(0, 0, 0, 0)
bg:SetDrawLayer("ARTWORK", 1)
BarFrame.bg = bg

local function createBar(parent)
	local bar = CreateFrame("Frame", nil, parent)
	bar:SetWidth(parent:GetWidth())
	bar:SetHeight(parent:GetHeight()/5)

	local fg = bar:CreateTexture(nil, "ARTWORK")
	fg:ClearAllPoints()
	fg:SetPoint("TOPLEFT")
	fg:SetPoint("BOTTOMLEFT")
	fg:SetColorTexture(unpack(redBar))
	fg:SetDrawLayer("ARTWORK", 2)
	bar.fg = fg

	local bg = bar:CreateTexture(nil, "ARTWORK")
	bg:ClearAllPoints()
	bg:SetAllPoints()
	bg:SetColorTexture(0, 0, 0, .5)
	bg:SetDrawLayer("ARTWORK", 1)
	bar.bg = bg

	local timer = bar:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
	timer:SetNonSpaceWrap(true)
	timer:SetPoint("RIGHT", -2, 0)
	bar.timer = timer

	local text = bar:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
	text:SetNonSpaceWrap(true)
	text:SetPoint("LEFT", 2, 0)
	bar.text = text

	local iconFrame = CreateFrame("Frame", nil, bar)
	iconFrame:SetWidth(bar:GetHeight())
	iconFrame:SetHeight(bar:GetHeight())
	iconFrame:SetPoint("RIGHT", bar, "LEFT")
	local icon = iconFrame:CreateTexture(nil, "OVERLAY")
	icon:ClearAllPoints()
	icon:SetAllPoints()
	icon:SetTexCoord(.08, .92, .08, .92) -- Strip the "borders"
	icon:SetTexture(defaultIcon)
	bar.iconFrame = iconFrame
	bar.icon = icon

	return bar
end

do
	BarFrame[1] = createBar(BarFrame)
	BarFrame[1]:SetPoint("TOPLEFT", BarFrame, "TOPLEFT")

	BarFrame[2] = createBar(BarFrame)
	BarFrame[2]:SetPoint("TOPLEFT", BarFrame[1], "BOTTOMLEFT")

	BarFrame[3] = createBar(BarFrame)
	BarFrame[3]:SetPoint("TOPLEFT", BarFrame[2], "BOTTOMLEFT")

	BarFrame[4] = createBar(BarFrame)
	BarFrame[4]:SetPoint("TOPLEFT", BarFrame[3], "BOTTOMLEFT")

	BarFrame[5] = createBar(BarFrame)
	BarFrame[5]:SetPoint("TOPLEFT", BarFrame[4], "BOTTOMLEFT")

	for i = 1, 5 do
		BarFrame[i].fg:SetWidth(BarFrame:GetWidth())
		BarFrame[i].fg:SetColorTexture(unpack(greenBar))
		BarFrame[i].timer:SetText("Ready")
	end
end

------------------------------------------------------------------------

local function getInterruptSpell(class, name)
	-- return spellId, cooldown
	if class == "DEATHKNIGHT" then
		return 47528, 15 -- Mind Freeze
	elseif class == "DEMONHUNTER" then
		return 183752, 15 -- Consume Magic
	elseif class == "DRUID" then
		return 106839, 15 -- Skull Bash
	elseif class == "HUNTER" then
		return 147362, 24 -- Counter Shot
	elseif class == "MAGE" then
		return 2139, 24 -- Counterspell
		--return 1953, 15 -- Blink
	elseif class == "MONK" then
		return 116705, 15 -- Spear Hand Strike
		--return 124081, 10 -- Zen Sphere
	elseif class == "PALADIN" then
		return 96231, 15 -- Rebuke
	elseif class == "PRIEST" then
		return 15487, 45 -- Silence
	elseif class == "ROGUE" then
		return 1766, 15 -- Kick
	elseif class == "SHAMAN" then
		return 57994, 12 -- Wind Shear
	elseif class == "WARLOCK" then
		--return 19647, 24 -- Spell Lock (Felhunter)
		return 119910, 171140, 24 -- Command Demon (Felhunter - Spell Lock), Command Demon (Doomguard - Shadow Lock)
	elseif class == "WARRIOR" then
		return 6552, 15 -- Pummel
	else 
		return 0, 0
	end
end

local function getViewCondition()
	if IsInGroup() then
		-- Raid or Party
		if IsInRaid() then
			return cfg.raid
		else
			return cfg.party
		end
	else
		-- Solo
		return cfg.solo
	end

	return false
end

local function updateBars()
	for i = 1, 5 do
		if Interrupters[i] and Interrupters[i].name and Interrupters[i].name ~= "" and Interrupters[i] ~= nil then
			BarFrame[i]:Show()
			BarFrame[i].text:SetText(Interrupters[i].name)
			BarFrame[i].icon:SetTexture(Interrupters[i].icon)
		else
			BarFrame[i]:Hide()
		end
	end
end

local function processName(pos, name)
	if Interrupters[pos] then
		wipe(Interrupters[pos])
	end
	if name and name ~= "" and name ~= nil and UnitClass(name) then
		local _, class = UnitClass(name)
		local spellId, spellId2, cooldown
		if class == "WARLOCK" then
			spellId, spellId2, cooldown = getInterruptSpell(class)
		else
			spellId, cooldown = getInterruptSpell(class)
		end
		local _, _, icon = GetSpellInfo(spellId)

		Debug("processName:", pos, name, class, spellId, cooldown, icon)
		Interrupters[pos] = {
			["name"] = name or "",
			["class"] = class or "",
			["spellId"] = spellId or 0,
			["spellId2"] = spellId2 or 0,
			["CD"] = cooldown or 0,
			["icon"] = icon or defaultIcon,
			["timer"] = 0,
			["alive"] = true,
		}
	end

	updateBars()
end

local function updateNames(data)
	if #data == 0 then
		Debug("! Empty data input.")
		f:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
		f:UnregisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
		f.SPELL_UPDATE_COOLDOWN = nil
		f.ACTIONBAR_UPDATE_COOLDOWN = nil
	end

	wipe(Interrupters)
	for pos, name in pairs(data) do
		if name and name ~= "" and name ~= nil and UnitClass(name) then
			local _, class = UnitClass(name)
			local spellId, spellId2, cooldown
			if class == "WARLOCK" then
				spellId, spellId2, cooldown = getInterruptSpell(class)
			else
				spellId, cooldown = getInterruptSpell(class)
			end
			local _, _, icon = GetSpellInfo(spellId)

			Debug("updateNames:", pos, name, class, spellId, cooldown, icon)
			Interrupters[pos] = {
				["name"] = name or "",
				["class"] = class or "",
				["spellId"] = spellId or 0,
				["spellId2"] = spellId2 or 0,
				["CD"] = cooldown or 0,
				["icon"] = icon or defaultIcon,
				["timer"] = 0,
				["alive"] = true,
			}
		end
	end

	if #Interrupters > 0 then
		Debug("- Interrupters:", #Interrupters)
		f:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
		f:UnregisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
		f.SPELL_UPDATE_COOLDOWN = nil
		f.ACTIONBAR_UPDATE_COOLDOWN = nil
	end
	updateBars()
end

local function getPartyNames()
	if #partyTable > 0 then return partyTable end

	for i = 0, GetNumGroupMembers() do
		local id = (i == 0 and "player") or "party"..i
		partyTable[i + 1] = UnitName(id)
	end

	return partyTable
end

------------------------------------------------------------------------

local function setupGroup(container)
	local header = AceGUI:Create("Heading")
	header:SetText("Interrupters")
	header:SetFullWidth(true)
	container:AddChild(header)

	local int1 = AceGUI:Create("EditBox")
	int1:SetLabel("1st interrupter:")
	int1:SetFullWidth(true)
	int1:SetText(db[1] or "")
	int1:DisableButton(false)
	int1:SetCallback("OnEnterPressed", function(widget, event, text)
		AceGUI:ClearFocus()
		db[1] = text
		processName(1, db[1])
		Debug("- int1", event, text)
	end)
	container:AddChild(int1)

	local int2 = AceGUI:Create("EditBox")
	int2:SetLabel("2nd interrupter:")
	int2:SetFullWidth(true)
	int2:SetText(db[2] or "")
	int2:DisableButton(false)
	int2:SetCallback("OnEnterPressed", function(widget, event, text)
		AceGUI:ClearFocus()
		db[2] = text
		processName(2, db[2])
		Debug("- int2", text)
	end)
	container:AddChild(int2)

	local int3 = AceGUI:Create("EditBox")
	int3:SetLabel("3rd interrupter:")
	int3:SetFullWidth(true)
	int3:SetText(db[3] or "")
	int3:DisableButton(false)
	int3:SetCallback("OnEnterPressed", function(widget, event, text)
		AceGUI:ClearFocus()
		db[3] = text
		processName(3, db[3])
		Debug("- int3", text)
	end)
	container:AddChild(int3)

	local int4 = AceGUI:Create("EditBox")
	int4:SetLabel("4th interrupter:")
	int4:SetFullWidth(true)
	int4:SetText(db[4] or "")
	int4:DisableButton(false)
	int4:SetCallback("OnEnterPressed", function(widget, event, text)
		AceGUI:ClearFocus()
		db[4] = text
		processName(4, db[4])
		Debug("- int4", text)
	end)
	container:AddChild(int4)

	local int5 = AceGUI:Create("EditBox")
	int5:SetLabel("5th interrupter:")
	int5:SetFullWidth(true)
	int5:SetText(db[5] or "")
	int5:DisableButton(false)
	int5:SetCallback("OnEnterPressed", function(widget, event, text)
		AceGUI:ClearFocus()
		db[5] = text
		processName(5, db[5])
		Debug("- int5", text)
	end)
	container:AddChild(int5)

	local button = AceGUI:Create("Button")
	button:SetText("Solo (Sending disabled)")
	button:SetWidth(200)
	button:SetCallback("OnClick", function()
		if not IsInGroup() then return end

		if UnitAffectingCombat("player") then
			Print("You are in combat, not sending the list.")
			return
		end

		for n = 1, 5 do -- Replace nils with empty strings
			if not db[n] or db[n] == nil then
				db[n] = ""
			end
		end
		local sync = strjoin(":", tostringall(unpack(db)))
		local commType = IsInRaid() and "RAID" or "PARTY"
		SendAddonMessage(ADDON_NAME, "L="..(sync or ""), commType)
		Print("Interrupt list sent to", strlower(commType))
		Debug("> Sending list:", sync, commType)

		SendChatMessage("Interrupters:", commType)
		for i = 1, 5 do
			if db[i] and db[i] ~= "" then
				SendChatMessage(i.." "..db[i], commType)
			end
		end
	end)
	if not IsInGroup() then
		button:SetDisabled(true)
	else
		button:SetDisabled(false)
		if IsInRaid() then
			button:SetText("Send to Raid")
		else
			button:SetText("Send to Party")
		end
	end
	container:AddChild(button)
end

local function configGroup(container)
	local showHead = AceGUI:Create("Heading")
	showHead:SetText("Show when")
	showHead:SetFullWidth(true)
	container:AddChild(showHead)

	local solo = AceGUI:Create("CheckBox")
	solo:SetValue(cfg.solo)
	solo:SetLabel("Solo")
	solo:SetCallback("OnValueChanged", function(widget, event, value)
		cfg.solo = value
		Debug("- solo", tostring(value))
	end)
	container:AddChild(solo)

	local party = AceGUI:Create("CheckBox")
	party:SetValue(cfg.party)
	party:SetLabel("in Party")
	party:SetCallback("OnValueChanged", function(widget, event, value)
		cfg.party = value
		Debug("- party", tostring(value))
	end)
	container:AddChild(party)

	local raid = AceGUI:Create("CheckBox")
	raid:SetValue(cfg.raid)
	raid:SetLabel("in Raid")
	raid:SetCallback("OnValueChanged", function(widget, event, value)
		cfg.raid = value
		Debug("- raid", tostring(value))
	end)
	container:AddChild(raid)

	local combatonly = AceGUI:Create("CheckBox")
	combatonly:SetValue(cfg.combatonly)
	combatonly:SetLabel("in combat only")
	combatonly:SetCallback("OnValueChanged", function(widget, event, value)
		cfg.combatonly = value
		Debug("- combatonly", tostring(value))
	end)
	container:AddChild(combatonly)

	local positionHead = AceGUI:Create("Heading")
	positionHead:SetText("Position")
	positionHead:SetFullWidth(true)
	container:AddChild(positionHead)

	local XSlider = AceGUI:Create("Slider")
	XSlider:SetFullWidth(true)
	XSlider:SetSliderValues(-floor(GetScreenWidth()/2 + 0.5), floor(GetScreenWidth()/2 + 0.5), 1)
	XSlider:SetValue(cfg.XPos)
	XSlider:SetLabel("X-position from CENTER")
	XSlider:SetCallback("OnMouseUp", function(widget, event, value)
		cfg.XPos = value
		BarFrame:SetPoint("CENTER", cfg.XPos, cfg.YPos)
		Debug("- XSlider", tostring(value))
	end)
	container:AddChild(XSlider)

	local YSlider = AceGUI:Create("Slider")
	YSlider:SetFullWidth(true)
	YSlider:SetSliderValues(-floor(GetScreenHeight()/2 + 0.5), floor(GetScreenHeight()/2 + 0.5), 1)
	YSlider:SetValue(cfg.YPos)
	YSlider:SetLabel("Y-position from CENTER")
	YSlider:SetCallback("OnMouseUp", function(widget, event, value)
		cfg.YPos = value
		BarFrame:SetPoint("CENTER", cfg.XPos, cfg.YPos)
		Debug("- YSlider", tostring(value))
	end)
	container:AddChild(YSlider)

	local announceHead = AceGUI:Create("Heading")
	announceHead:SetText("Announce")
	announceHead:SetFullWidth(true)
	container:AddChild(announceHead)

	local announce = AceGUI:Create("CheckBox")
	announce:SetFullWidth(true)
	announce:SetValue(cfg.announce)
	announce:SetLabel("Announce interrupts (when in group)")
	announce:SetCallback("OnValueChanged", function(widget, event, value)
		cfg.announce = value
		Debug("- announce", tostring(value))
	end)
	container:AddChild(announce)

	local resetHead = AceGUI:Create("Heading")
	resetHead:SetText("Reset settings and interrupter list")
	resetHead:SetFullWidth(true)
	container:AddChild(resetHead)

	local button = AceGUI:Create("Button")
	button:SetText("Hold Shift and Click")
	button:SetWidth(200)
	button:SetCallback("OnClick", function()
		if IsShiftKeyDown() then
			Debug("! Reset settings")
			wipe(cfg)
			wipe(db)
			initDB(cfg, defaultDB)
			initDB(db)

			Debug("- Adding player to the list")
			db[1] = playerName
			if cfg.partyFill and IsInGroup() and not IsInRaid() then
				updateNames(getPartyNames())
			else
				updateNames(db)
			end

			BarFrame:SetPoint("CENTER", cfg.XPos, cfg.YPos)
			tab:SelectTab("config")
		end
		Debug("- Reset", tostring(IsShiftKeyDown()))
	end)
	container:AddChild(button)
end

local function SelectGroup(container, event, group)
	container:ReleaseChildren()
	if group == "setup" then
		setupGroup(container)
		config:SetWidth(250)
		--tab:SetTitle("Setup - These are saved per character.")
		openSetup = true
	elseif group == "config" then
		configGroup(container)
		config:SetWidth(500)
		--tab:SetTitle("Config - These are shared between all charaters.")
		openSetup = false
	end
end

local function createCfg()
	config = AceGUI:Create("Frame")
	config:SetWidth(275)
	config:SetHeight(400)
	config:EnableResize(false)
	config:SetTitle(ADDON_NAME)
	config:SetCallback("OnClose", function(widget)
		AceGUI:Release(widget)
		BarFrame.bg:SetColorTexture(0, 0, 0, 0)
		openSetup = false
		if not getViewCondition() or cfg.combatonly then
			BarFrame:Hide()
		end
	end)
	config:SetLayout("Fill")

	tab = AceGUI:Create("TabGroup")
	tab:SetLayout("Flow")
	tab:SetTabs({ { text = "Setup", value = "setup" }, { text = "Config", value = "config" } })
	tab:SetCallback("OnGroupSelected", SelectGroup)
	tab:SelectTab("setup")
	config:AddChild(tab)

	BarFrame.bg:SetColorTexture(0, 0, 0, .5)
	BarFrame:Show()
end

------------------------------------------------------------------------

SLASH_LOIHINTERRUPT1 = "/lint"
SLASH_LOIHINTERRUPT2 = "/loihint"

local SlashHandlers = {
	["debug"] = function()
		local setting = not cfg.debugmode
		cfg.debugmode = setting
		Print("debugmode:", cfg.debugmode and "|cff00ff00true|r" or "|cffff0000false|r")
	end,
	["show"] = function()
		BarFrame:Show()
	end,
	["hide"] = function()
		BarFrame:Hide()
	end,
	["config"] = function()
		createCfg()
	end,
	["reset"] = function()
		Debug("! Reset settings")
		wipe(cfg)
		wipe(db)
		initDB(cfg, defaultDB)
		initDB(db)

		Debug("- Adding player to the list")
		db[1] = playerName
		if cfg.partyFill and IsInGroup() and not IsInRaid() then
			updateNames(getPartyNames())
		else
			updateNames(db)
		end

		BarFrame:SetPoint("CENTER", cfg.XPos, cfg.YPos)
		BarFrame:Show()

		-- Not needed but just in case if I change defaults in the future
		if getViewCondition() then
			if cfg.combatonly then
				BarFrame:Hide()
			end
		elseif not (config and config:IsVisible()) then
			BarFrame:Hide()
		end
	end,
}

SlashCmdList["LOIHINTERRUPT"] = function(text)
	local command, params = strsplit(" ", text, 2)

	if UnitAffectingCombat("player") and not (command == "show" or command == "hide") then
		Print("You are in combat, your command is processed after you leave combat.")
		combatslash = (text ~= "" and text ~= nil) and text or "config"
		return
	end

	if SlashHandlers[command] then
		SlashHandlers[command](params)
	else
		createCfg()
	end
end

------------------------------------------------------------------------

function f:ADDON_LOADED(event, addon)
	if addon ~= ADDON_NAME then return end
	self:UnregisterEvent(event)

	LOIHInterruptDB = initDB(LOIHInterruptDB, defaultDB)
	LOIHInterruptCharDB = initDB(LOIHInterruptCharDB)
	cfg = LOIHInterruptDB
	db = LOIHInterruptCharDB

	BarFrame:SetPoint("CENTER", cfg.XPos, cfg.YPos)

	if IsLoggedIn() then
		self:PLAYER_LOGIN()
	else
		self:RegisterEvent("PLAYER_LOGIN")
	end

	self.ADDON_LOADED = nil
end

function f:PLAYER_LOGIN(event)
	self:UnregisterEvent(event)

	RegisterAddonMessagePrefix(ADDON_NAME)
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("GROUP_ROSTER_UPDATE")

	playerName = UnitName("player")
	playerClass = select(2, UnitClass("player"))

	-- Let's use these two events to get updated interrupt spell data
	-- when name(s) is/are used as parameter for UnitClass-function.
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")

	if cfg.partyFill and IsInGroup() and not IsInRaid() then
		updateNames(getPartyNames())
	elseif #db > 0 then
		updateNames(db)
	else
		Debug("- Adding player to the list")
		db[1] = playerName
		updateNames(db)
	end

	self.PLAYER_LOGIN = nil
end

function f:SPELL_UPDATE_COOLDOWN(event)
	if cfg.partyFill and IsInGroup() and not IsInRaid() then
		updateNames(getPartyNames())
	else
		updateNames(db)
	end
end

function f:ACTIONBAR_UPDATE_COOLDOWN(event)
	if cfg.partyFill and IsInGroup() and not IsInRaid() then
		updateNames(getPartyNames())
	else
		updateNames(db)
	end
end

function f:CHAT_MSG_ADDON(event, prefix, message, channel, sender)
	--prefix == ADDON_NAME and Debug("< Message:", prefix, message, channel, sender)
	if not ((channel == "RAID" or channel == "PARTY") and prefix == ADDON_NAME) then return end
	--if not ((channel == "RAID" or channel == "PARTY" or channel == "GUILD") and prefix == ADDON_NAME) then return end -- Debug

	local shortSender = Ambiguate(sender, "none")
	local command, data = strsplit("=", message)
	if command == "L" and shortSender ~= playerName then
	--if command == "L" then -- Debug
		db = { strsplit(":", data) }
		if not cfg.partyFill then
			updateNames(db)
		end

		if config and config:IsVisible() and openSetup then -- Update names if Setup is open when Sync comes in
			tab:SelectTab("setup")
		end

		Print("Interrupt list recieved from", shortSender)
		Debug("< Received list:", data, "from", shortSender)
	end
end

function f:PLAYER_ENTERING_WORLD(event)
	self:UnregisterEvent(event)

	if getViewCondition() then
		if cfg.combatonly then
			BarFrame:Hide()
		end
	elseif not (config and config:IsVisible()) then
		BarFrame:Hide()
	end

	self.PLAYER_ENTERING_WORLD = nil
end

function f:PLAYER_REGEN_DISABLED(event)
	self:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
	self:UnregisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
	self.SPELL_UPDATE_COOLDOWN = nil
	self.ACTIONBAR_UPDATE_COOLDOWN = nil


	if cfg.partyFill and IsInGroup() and not IsInRaid() then
		updateNames(getPartyNames())
	else
		updateNames(db)
	end

	if getViewCondition() then
		if cfg.combatonly then
			BarFrame:Show()
		end
	elseif not (config and config:IsVisible()) then
		BarFrame:Hide()
	end
end

function f:PLAYER_REGEN_ENABLED(event)
	if cfg.combatonly and not (config and config:IsVisible()) then
		BarFrame:Hide()
	end

	-- Slash command used during combat, wait until you leave combat
	if combatslash then
		local command, params = strsplit(" ", combatslash, 2)

		if SlashHandlers[command] then
			SlashHandlers[command](params)
		else
			createCfg()
		end
		combatslash = false
	end
end

function f:GROUP_ROSTER_UPDATE(event)
	Debug(event)
	wipe(partyTable)
	if cfg.partyFill and IsInGroup() and not IsInRaid() then
		updateNames(getPartyNames())
	else
		updateNames(db)
	end

	if getViewCondition() then
		if not cfg.combatonly then
			BarFrame:Show()
		end
	elseif not (config and config:IsVisible()) then
		BarFrame:Hide()
	end

	if config and config:IsVisible() and openSetup then -- Update the Sync-button text if Setup is open when changing between solo, party and raid
		tab:SelectTab("setup")
	end
end

function f:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	--local timestamp, eventType, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellId, spellName, spellSchool = ...
	--local _, eventType, _, _, sourceName, _, _, _, destName, _, _, spellId, spellName = ...
	local _, eventType, _, _, sourceName, _, _, _, destName, _, _, spellId, spellName, _, extraSpellId, extraSpellName = ...

	-- CD bars
	if eventType == "SPELL_CAST_SUCCESS" then
		for i = 1, 5 do
			if Interrupters[i] and sourceName == Interrupters[i].name and (spellId == Interrupters[i].spellId or spellId == Interrupters[i].spellId2) then
				Interrupters[i].timer = Interrupters[i].CD
				break
			end
		end
	end

	-- Announce
	if cfg.announce and IsInGroup() and eventType == "SPELL_INTERRUPT" and sourceName == playerName and (spellId == (getInterruptSpell(playerClass)) or spellId == select(2, { getInterruptSpell(playerClass) })) then -- Dirty WARLOCKs made me use 'select'
	--if eventType == "SPELL_INTERRUPT" and sourceName == playerName and spellId == getInterruptSpell(playerClass) then -- Debug
		SendChatMessage(format(interruptMsg, extraSpellId, extraSpellName), "SAY")
	end

	if eventType == "UNIT_DIED" then
		for i = 1, 5 do
			if Interrupters[i] and destName == Interrupters[i].name then
				Interrupters[i].alive = false
				deadWatch[i] = Interrupters[i].name
			end
		end
	end
	-- SPELL_RESURRECT
		-- timestamp, SPELL_AURA_APPLIED, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, 160029, "Resurrecting", spellSchool, DEBUFF
end

local function OnUpdate(self, elapsed)
	local now = GetTime()

	local ceilTime = math.ceil(now)
	if deadCheck ~= ceilTime then -- Check this only once per second
		deadCheck = ceilTime

		for k, v in pairs(deadWatch) do -- Check if dead people have been Ressurected or some other way brought back to life
			if k and v and UnitHealth(v) > 0 then
				Interrupters[k].alive = true
				deadWatch[k] = nil
			end
		end
	end

	for i = 1, 5 do
		if Interrupters[i] and Interrupters[i].timer ~= nil and Interrupters[i].timer > 0 then -- Was on CD
			Interrupters[i].timer = Interrupters[i].timer - elapsed

			if Interrupters[i].timer > 0 then -- Still on CD
				local progress = Interrupters[i].timer / math.max(Interrupters[i].CD, 1)
				BarFrame[i].fg:SetWidth(BarFrame:GetWidth() * progress)
				BarFrame[i].timer:SetFormattedText("%.1f", Interrupters[i].timer)
				if Interrupters[i].alive == true then -- Alive
					BarFrame[i].fg:SetColorTexture(unpack(redBar))
				else -- Dead
					BarFrame[i].fg:SetColorTexture(unpack(blackBar))
				end
			else -- Not on CD anymore
				BarFrame[i].fg:SetWidth(BarFrame:GetWidth())
				if Interrupters[i].alive == true then -- Alive
					BarFrame[i].fg:SetColorTexture(unpack(greenBar))
					BarFrame[i].timer:SetText("Ready")
				else -- Dead
					BarFrame[i].fg:SetColorTexture(unpack(blackBar))
					BarFrame[i].timer:SetText("Dead")
				end
			end
		elseif Interrupters[i] then -- Was Ready
			BarFrame[i].fg:SetWidth(BarFrame:GetWidth())
			if Interrupters[i].alive == true then -- Alive
				BarFrame[i].fg:SetColorTexture(unpack(greenBar))
				BarFrame[i].timer:SetText("Ready")
			else -- Dead
				BarFrame[i].fg:SetColorTexture(unpack(blackBar))
				BarFrame[i].timer:SetText("Dead")
			end
		end
	end
end
f:SetScript("OnUpdate", OnUpdate)